@@ -18,6 +18,8 @@ import logging
 import wx
 import wx.lib.newevent as wxevent
 
+import fsleyes_widgets.utils as wutils
+
 
 log = logging.getLogger(__name__)
 
@@ -29,20 +31,22 @@ class AutoTextCtrl(wx.Panel):
     """
 
 
-    def __init__(self, parent, style=0, ownloop=True):
+    def __init__(self, parent, style=0, modal=True):
         """Create an ``AutoTextCtrl``.
 
         :arg parent: The ``wx`` parent object.
         :arg style:  Can be :data:`ATC_CASE_SENSITIVE` to restrict the
                      auto-completion options to case sensitive matches.
-        arg ownloop: See :meth:`AutoCompletePopup.__init__`.
+        :arg modal:  If ``True`` (the default), the :class:`AutoCompletePopup`
+                     is shoown modally. This option is primarily for testing
+                     purposes.
         """
 
-        self.__ownloop       = ownloop
         self.__caseSensitive = style & ATC_CASE_SENSITIVE
 
         wx.Panel.__init__(self, parent)
 
+        self.__modal    = modal
         self.__popup    = None
         self.__textCtrl = wx.TextCtrl(self, style=wx.TE_PROCESS_ENTER)
         self.__sizer    = wx.BoxSizer(wx.HORIZONTAL)
@@ -223,8 +227,7 @@ class AutoTextCtrl(wx.Panel):
             self,
             text,
             self.__options,
-            style,
-            self.__ownloop)
+            style)
 
         if popup.GetCount() == 0:
             popup.Destroy()
@@ -261,7 +264,9 @@ class AutoTextCtrl(wx.Panel):
 
         popup.SetSize((-1, -1))
         popup.SetPosition((posx,  posy))
-        popup.Show()
+
+        if self.__modal: popup.ShowModal()
+        else:            popup.Show()
 
 
 ATC_CASE_SENSITIVE = 1
@@ -286,12 +291,12 @@ Contains a single attribute, ``text``, which contains the text in the
 """
 
 
-class AutoCompletePopup(wx.Frame):
+class AutoCompletePopup(wx.Dialog):
     """The ``AutoCompletePopup`` class is used by the :class:`AutoTextCtrl`
     to display a list of completion options to the user.
     """
 
-    def __init__(self, parent, atc, text, options, style=0, ownloop=True):
+    def __init__(self, parent, atc, text, options, style=0):
         """Create an ``AutoCompletePopup``.
 
         :arg parent:  The ``wx`` parent object.
@@ -300,20 +305,13 @@ class AutoCompletePopup(wx.Frame):
         :arg options: A list of all possible auto-completion options.
         :arg style:   Set to :data:`ATC_CASE_SENSITIVE` to make the
                       pattern matching case sensitive.
-        :arg ownloop: If ``True`` (the default), a separate event loop
-                      is created and run while the popup is displayed.
-                      This gives the popup modal behaviour.
         """
 
-        wx.Frame.__init__(self,
+        wx.Dialog.__init__(self,
                           parent,
-                          style=(wx.NO_BORDER        |
-                                 wx.STAY_ON_TOP      |
-                                 wx.FRAME_NO_TASKBAR |
-                                 wx.FRAME_FLOAT_ON_PARENT))
+                          style=(wx.NO_BORDER | wx.STAY_ON_TOP))
 
-        self.__ownloop       = None
-        self.__evloop        = None
+        self.__alive         = True
         self.__caseSensitive = style & ATC_CASE_SENSITIVE
         self.__atc           = atc
         self.__options       = options
@@ -367,25 +365,6 @@ class AutoCompletePopup(wx.Frame):
         self.__listBox .Bind(wx.EVT_SET_FOCUS,     self.__onSetFocus)
 
 
-    def Show(self):
-        """Shows this ``AutoCompletePopup``. If the ``ownloop`` parameter
-        passed to :meth:`__init__` was ``True``, a A new ``wx.GUIEventLoop``
-        is created and started, until this ``AutoCompletePopup`` loses focus.
-        """
-
-        parent = self.GetParent().GetTopLevelParent()
-
-        parent.Disable()
-        wx.Frame.Show(self)
-
-        if self.__ownloop:
-            self.__evloop = wx.GUIEventLoop()
-            self.__evloop.Run()
-            self.__evloop = None
-
-        parent.Enable()
-
-
     def GetCount(self):
         """Returns the number of auto-completion options currently available.
         """
@@ -443,6 +422,14 @@ class AutoCompletePopup(wx.Frame):
         this ``AutoCompletePopup`` to the owning :class:`AutoTextCtrl`,
         and then (asynchronously) destroys this ``AutoCompletePopup``.
         """
+
+        # destroy has already been
+        # called - don't run again
+        if not self.__alive:
+            return
+
+        self.__alive = False
+
         value = self.__textCtrl.GetValue()
         idx   = self.__textCtrl.GetInsertionPoint()
         atc   = self.__atc
@@ -451,12 +438,21 @@ class AutoCompletePopup(wx.Frame):
         # events, which will trigger another call to
         # __destroy. So we remove all callbacks to
         # prevent this from happening.
-        self           .Bind(wx.EVT_SET_FOCUS,  None)
-        self.__textCtrl.Bind(wx.EVT_SET_FOCUS,  None)
-        self.__listBox .Bind(wx.EVT_SET_FOCUS,  None)
-        self           .Bind(wx.EVT_KILL_FOCUS, None)
-        self.__textCtrl.Bind(wx.EVT_KILL_FOCUS, None)
-        self.__listBox .Bind(wx.EVT_KILL_FOCUS, None)
+        self.__textCtrl.Bind(wx.EVT_TEXT,           None)
+        self.__textCtrl.Bind(wx.EVT_TEXT_ENTER,     None)
+        self.__textCtrl.Bind(wx.EVT_CHAR_HOOK,      None)
+        self.__textCtrl.Bind(wx.EVT_KEY_DOWN,       None)
+        self.__listBox .Bind(wx.EVT_CHAR_HOOK,      None)
+        self.__listBox .Bind(wx.EVT_KEY_DOWN,       None)
+        self.__listBox .Bind(wx.EVT_LISTBOX_DCLICK, None)
+        self.__listBox .Bind(wx.EVT_LEFT_DOWN,      None)
+        self.__listBox .Bind(wx.EVT_RIGHT_DOWN,     None)
+        self           .Bind(wx.EVT_SET_FOCUS,      None)
+        self.__textCtrl.Bind(wx.EVT_SET_FOCUS,      None)
+        self.__listBox .Bind(wx.EVT_SET_FOCUS,      None)
+        self           .Bind(wx.EVT_KILL_FOCUS,     None)
+        self.__textCtrl.Bind(wx.EVT_KILL_FOCUS,     None)
+        self.__listBox .Bind(wx.EVT_KILL_FOCUS,     None)
 
         atc.ChangeValue(      value)
         atc.SetInsertionPoint(idx)
@@ -470,23 +466,19 @@ class AutoCompletePopup(wx.Frame):
             atc.GenEnterEvent()
 
         def destroy():
-            try:
+
+            if not wutils.isalive(self):
+                return
+
+            if self.IsModal():
+                self.EndModal(wx.ID_OK)
+            else:
                 self.Close()
                 self.Destroy()
 
-            except wx.PyDeadObjectError:
-                pass
-
         ev = ATCPopupDestroyEvent()
         ev.SetEventObject(self)
         wx.PostEvent(self, ev)
-
-        # Kill the event loop used by
-        # the popup (see the Show method)
-        if self.__evloop is not None:
-            self.__evloop.Exit()
-            self.__evloop = None
-
         wx.CallAfter(destroy)
 
 
